from pydantic import BaseSettings

import pathlib

_SETTINGS = None


class Settings(BaseSettings):
    PROG_HOME: pathlib.Path = pathlib.Path("~").expanduser() / ".tchotcho"
    GPU_INFO_FILE: pathlib.Path = PROG_HOME / "gpu_info.json"
    PREFIX: str = "tchotcho"

    # class Config:
    #     env_file = os.environ.get("TCHOTCHO_ENV", ".env")


def set_settings(settings):
    global _SETTINGS

    if not settings.PROG_HOME.exists():
        settings.PROG_HOME.mkdir()

    _SETTINGS = settings
    return _SETTINGS


def get_settings():
    return _SETTINGS


set_settings(Settings())



import structlog
import os
from logma.wech import datlog

LOGGER_NAME = "tchotcho"
LEVEL = os.getenv("TCHOTCHO_LOG_LEVEL", "INFO")
LEVEL_EXTERN = os.getenv("TCHOTCHO_LOG_LEVEL", "ERROR")

config = {
    "loggers": {
        LOGGER_NAME: {"handlers": ["default"], "level": LEVEL, "propagate": False}
    }
}

tty = os.getenv("TCHOTCHO_LOG_TTY", "")
tty_flag = None
if tty:
    tty_flag = tty.lower() in ("an", "on", "true", "1")
# XXX all other loggers are on level ERROR e.g. boto, ...
datlog(level=LEVEL_EXTERN, tty=tty_flag, user_config=config)
log = structlog.get_logger(LOGGER_NAME)



from tchotcho.action.key import KeyManager  # noqa
from tchotcho.action.stack import StackManager  # noqa
from tchotcho.action.spot import SpotManager  # noqa
from tchotcho.action.info import InfoManager  # noqa
from tchotcho.action.shell import ShellManager  # noqa



import hashlib

from Crypto.PublicKey import RSA


class RSAFingerprintManager(object):
    def to_string(self, string):
        return ":".join(a + b for a, b in zip(string[::2], string[1::2]))

    def get_rsa_key(self, key_location, passphrase):
        with open(key_location) as f:
            key = RSA.importKey(f.read(), passphrase=passphrase)
        return key

    def get_private(self, key_location, passphrase=None):
        k = self.get_rsa_key(key_location, passphrase)
        sha1digest = hashlib.sha1(k.exportKey("DER", pkcs=8)).hexdigest()
        fingerprint = self.to_string(sha1digest)
        return fingerprint

    def get_public(self, key_location, passphrase=None):
        privkey = self.get_rsa_key(key_location, passphrase)
        pubkey = privkey.publickey()
        md5digest = hashlib.md5(pubkey.exportKey("DER")).hexdigest()
        fingerprint = self.to_string(md5digest)
        return fingerprint

    def extract_private_to_public(self, key_location, passphrase=None):
        # ssh-keygen -f ~/.ssh/test-dl.pem -y > ~/.ssh/test-dl.pem.pub
        privkey = self.get_rsa_key(key_location, passphrase)
        public_key = privkey.publickey().exportKey('PEM')
        return public_key.decode("ascii")



import json
import itertools
import concurrent.futures
import pandas as pd

import boto3
import click
import colorama
import requests

import tabulate

from tchotcho.config import get_settings

colorama.init()


class InfoManager(object):
    def __init__(self):
        self.settings = get_settings()

    def _get_ami(self, region, ownerid, namefilter, limit):
        rname = region["RegionName"]
        # print("Running query in region: %s" % rname)
        ec2 = boto3.client("ec2", region_name=rname)
        resp = ec2.describe_images(
            Owners=[ownerid], Filters=[{"Name": "name", "Values": [namefilter]}],
        )
        ret = []
        for image in resp["Images"]:
            ret.append((rname, image["Name"], image["CreationDate"], image["ImageId"]))
        # XXX return the newest <limit>
        ret = sorted(ret, key=lambda x: int(x[2].replace("-", "")[:8]), reverse=True)[
            :limit
        ]
        ret = [
            {"region": x[0], "name": x[1], "date": x[2], "ami": x[3]} for x in ret if x
        ]
        return ret

    def _get_ami_wrapper(self, args):
        return self._get_ami(*args)

    def get_ami_image(self, ownerid, namefilter, limit, method="future-process"):
        ec2 = boto3.client("ec2")
        response = ec2.describe_regions()
        regions = [(x, ownerid, namefilter, limit) for x in response["Regions"]]

        if method == "future-process":
            with concurrent.futures.ProcessPoolExecutor() as executor:
                data = list(executor.map(self._get_ami_wrapper, regions))
        else:
            data = [self._get_ami(*x) for x in regions]

        ret = list(itertools.chain.from_iterable(data))
        return ret

    def get_gpu_info(self):
        """
        Fetch gpu info from ec2instances.info
        """
        res = requests.get(
            "https://raw.githubusercontent.com/powdahound/ec2instances.info/master/www/instances.json"  # noqa
        )

        ret = []
        inst = res.json()

        for i in inst:
            # XXX has a gpu and g2 are old not supported gpu cards
            supported = i["GPU"] > 0 and not i["instance_type"].startswith("g2.")
            tmp = {
                "name": i["instance_type"],
                "gpu": i["GPU"],
                "cpu": i["vCPU"],
                "memory": i["memory"],
                "gpu_memory": i["GPU_memory"],
                "gpu_model": i["GPU_model"],
                "storage": i["storage"],
                "supported": supported,
            }

            pricing = {}
            for k in i["pricing"]:
                price = i["pricing"][k].get("linux", {}).get("ondemand")
                pricing[k] = float(price) if price else None
            tmp["pricing"] = pricing
            ret.append(tmp)
        return ret

    def update(self, ownerid, namefilter, limit):
        full = {}
        gpu_data = self.get_gpu_info()
        ami_data = self.get_ami_image(ownerid, namefilter, limit)
        full["instance"] = gpu_data
        full["ami"] = ami_data

        with open(self.settings.GPU_INFO_FILE, "w") as f:
            json.dump(full, f, indent=4)
        return full

    def list(self):
        with open(self.settings.GPU_INFO_FILE, "r") as f:
            return json.load(f)

    def render(self, data, region, csv):
        ami = data["ami"]
        gp = data["instance"]

        df = pd.DataFrame(gp)
        df_ami = pd.DataFrame(ami)
        df_ami = df_ami[df_ami.region == region]

        def set_price(val):
            val = val.get(region)
            val = float(val) if val else val
            return val

        df["price"] = df["pricing"].apply(set_price)

        def set_color(val):
            if str(val) != "nan":
                if val < 1:
                    val = colorama.Back.GREEN + str(val) + colorama.Back.RESET
                elif val > 3:
                    val = colorama.Back.RED + str(val) + colorama.Back.RESET
                else:
                    val = colorama.Back.YELLOW + str(val) + colorama.Back.RESET
            else:
                val = "Not available"
            return val

        def set_color_support(val):
            if val:
                val = colorama.Back.GREEN + str(val) + colorama.Back.RESET
            else:
                val = colorama.Back.YELLOW + str(val) + colorama.Back.RESET
            return val

        # apply to specific column
        df = df[
            [
                "name",
                "gpu",
                "gpu_memory",
                "gpu_model",
                "cpu",
                "memory",
                "supported",
                "price",
            ]
        ]
        df = df.sort_values(by=["gpu"], ascending=False)
        to_print = df.to_csv()
        if not csv:
            df["price"] = df["price"].apply(set_color)
            df["supported"] = df["supported"].apply(set_color_support)
            to_print = tabulate.tabulate(
                df, headers="keys", tablefmt="fancy_grid", showindex="never"
            )
        print(to_print)

        # ami
        to_print = df_ami.to_csv()
        if not csv:
            to_print = tabulate.tabulate(
                df_ami, headers="keys", tablefmt="fancy_grid", showindex="never"
            )
        print(to_print)


mgr = None


@click.group()
def info():
    global mgr
    mgr = InfoManager()


@info.command()
@click.option(
    "--ownerid",
    default="898082745236",
    required=True,
    help="Owner id used (we use ubuntu)",
    show_default=True,
)
@click.option(
    "--namefilter",
    default="Deep Learning AMI* 18.04*",
    show_default=True,
    help="AMI filter by name",
    required=True,
)
@click.option(
    "--limit",
    default=5,
    type=int,
    show_default=True,
    required=True,
    help="Number of AMI image",
)
@click.option("--csv/--no-csv", default=False)
@click.option("--region", help="List in region", required=True, default="eu-central-1")
def update(ownerid, namefilter, limit, csv, region):
    """Update the GPU info json file"""
    ret = mgr.update(ownerid, namefilter, limit)
    mgr.render(ret, region, csv)


@info.command(name="list")
@click.option("--region", help="List spot prices in region", required=True)
@click.option("--csv/--no-csv", default=False)
def _list(region, csv):
    """Get the GPU info"""
    ret = mgr.list()
    mgr.render(ret, region, csv)



import os

import boto3
import click
import colorama
import pandas as pd

import tabulate

from tchotcho.log import log
from tchotcho.rsa import RSAFingerprintManager
from tchotcho.config import get_settings

colorama.init()


class KeyManager:
    def __init__(self):
        self.ec2 = boto3.client("ec2")
        self.settings = get_settings()

    def _import(self, name, path):
        ret = False
        with open(path, "rb") as f:
            data = f.read()
        if self.exists(name):
            log.error(f"Key {name} already exists!")
            return ret

        resp = self.ec2.import_key_pair(KeyName=name, PublicKeyMaterial=data)
        if resp["ResponseMetadata"]["HTTPStatusCode"] == 200:
            log.info(f"Succesfully imported key {name} from {path}")
            ret = True
        return ret

    def list(self):
        resp = self.ec2.describe_key_pairs()  # KeyName=name)
        ret = [x for x in resp["KeyPairs"]]
        return ret

    def exists(self, name):
        return name in [x["KeyName"] for x in self.list()]

    def create(self, name):
        ret = None
        if self.exists(name):
            log.error(f"Key {name} already exists!")
            return None

        log.info(f"Creating key {name}...")
        resp = self.ec2.create_key_pair(KeyName=name)
        priv_key = resp["KeyMaterial"]

        pub_name = "%s.pub" % name
        PRIVATE_KEY_FILE = self.settings.PROG_HOME / name
        PUBLIC_KEY_FILE = self.settings.PROG_HOME / pub_name

        if resp["ResponseMetadata"]["HTTPStatusCode"] == 200:
            with open(PRIVATE_KEY_FILE, "w") as f:
                f.write(priv_key)
            os.chmod(PRIVATE_KEY_FILE, 0o400)
            log.info(f"Succesfully created private key {name} in {PRIVATE_KEY_FILE}")

            rsamgr = RSAFingerprintManager()
            pub_key = rsamgr.extract_private_to_public(PRIVATE_KEY_FILE)

            with open(PUBLIC_KEY_FILE, "w") as f:
                f.write(pub_key)
            os.chmod(PUBLIC_KEY_FILE, 0o400)
            log.info(f"Succesfully created public key {pub_name} in {PUBLIC_KEY_FILE}")

            ret = (PRIVATE_KEY_FILE, PUBLIC_KEY_FILE)

        return ret

    def delete(self, name, ask):
        ret = False
        if not self.exists(name):
            log.error(f"Key {name} does not exist!")
            return ret

        log.info(f"Deleting key {name}...")
        resp = self.ec2.delete_key_pair(KeyName=name)
        if resp["ResponseMetadata"]["HTTPStatusCode"] == 200:

            pub_name = "%s.pub" % name
            PRIVATE_KEY_FILE = self.settings.PROG_HOME / name
            PUBLIC_KEY_FILE = self.settings.PROG_HOME / pub_name
            for KEY_FILE in (PRIVATE_KEY_FILE, PUBLIC_KEY_FILE):
                if KEY_FILE.exists():
                    cond = input(f"Delete {KEY_FILE} locally? (yes/no): ") if ask else "yes"
                    if cond.strip().lower() == "yes":
                        KEY_FILE.unlink()
                log.info(f"Succesfully deleted key {name}!")
            ret = True
        return ret

    def fingerprint(self, path, passphrase):
        rsamgr = RSAFingerprintManager()
        ret = {
            "public": [rsamgr.get_public(path, passphrase)],
            "private": [rsamgr.get_private(path, passphrase)],
        }
        return ret


mgr = None


@click.group()
def key():
    global mgr
    mgr = KeyManager()


@key.command()
@click.option("--name", help="Name of key to highlight")
@click.option("--csv/--no-csv", default=False)
def list(name, csv):
    ret = mgr.list()

    def set_color(val):
        if val == name:
            val = colorama.Back.GREEN + val + colorama.Back.RESET
        return val

    # apply to specific column
    df = pd.DataFrame(ret)
    df = df[["KeyName", "KeyFingerprint"]]
    to_print = df.to_json()

    if not csv:
        df["KeyName"] = df["KeyName"].apply(set_color)
        to_print = tabulate.tabulate(
            df, headers="keys", tablefmt="fancy_grid", showindex="never"
        )
    print(to_print)


@key.command()
@click.option("--name", help="Name of key to create", required=True)
def create(name):
    mgr.create(name)
    click.echo(f"Succesfully created key {name}")


@key.command()
@click.option("--name", help="Name of key to highlight", required=True)
@click.option("--ask/--no-ask", default=True)
def delete(name, ask):
    ret = mgr.delete(name, ask)
    if ret:
        click.echo(f"Succesfully deleted key {name}")


@key.command(name="import")
@click.option("--name", help="Name of key to highlight", required=True)
@click.option(
    "--path",
    help="Path to key file",
    required=True,
    type=click.Path(exists=True, resolve_path=True),
)
def _import(name, path):
    ret = mgr._import(name, path)
    if ret:
        click.echo(f"Succesfully imported key {name}")


@key.command()
@click.option(
    "--path",
    help="Path to key file",
    required=True,
    type=click.Path(exists=True, resolve_path=True),
)
@click.option("--passphrase", prompt=True, hide_input=True, default="")
@click.option("--csv/--no-csv", default=False)
def fingerprint(path, passphrase, csv):
    ret = mgr.fingerprint(path, passphrase)

    # apply to specific column
    df = pd.DataFrame(ret)
    to_print = df.to_json()

    if not csv:
        to_print = tabulate.tabulate(
            df, headers="keys", tablefmt="fancy_grid", showindex="never"
        )
    print(to_print)



from tchotcho.log import log
import botocore


def boto_exception(fn):
    """Handle boto errors"""

    def decorator(self, *args, **kwargs):
        res = None
        try:
            res = fn(self, *args, **kwargs)
        except botocore.exceptions.ClientError as ex:
            error_message = ex.response["Error"]["Message"]
            log.exception("Boto error: %s" % error_message)
        except Exception:
            log.exception(f"An error occured on: {fn.__name__}")
        return res

    return decorator


def get_wrapped_waiter(cf, name, callback):
    waiter = cf.get_waiter(name)
    orig_func = waiter._operation_method

    def wrapper(**kwargs):
        response = orig_func(**kwargs)
        callback(response)
        return response

    waiter._operation_method = wrapper
    return waiter



import setuptools

with open("requirements.txt") as f:
    requirements = [x for x in f.read().split("\n") if x and not x.startswith("#")]


with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="tchotcho",
    version="0.6.0",
    author="Josip Delic",
    description="tchotcho",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/delijati/tchotcho",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
    ],
    zip_safe=False,
    include_package_data=True,
    package_dir={"": "src"},
    packages=setuptools.find_namespace_packages(where="src"),
    install_requires=requirements,
    extras_require={"tests": ["pytest", "pytest-cov", "pytest-env", "moto"]},
    entry_points={"console_scripts": ["tchotcho = tchotcho.__main__:cli"]},
)
